// Copyright 2001  EPITA Research and Development Laboratory
//
// This file is part of the Olena Library.  This library is free
// software; you can redistribute it and/or modify it under the terms
// of the GNU General Public License version 2 as published by the
// Free Software Foundation.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this library; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// As a special exception, you may use this file as part of a free
// software library without restriction.  Specifically, if other files
// instantiate templates or use macros or inline functions from this
// file, or you compile this file and link it with other files to
// produce an executable, this file does not by itself cause the
// resulting executable to be covered by the GNU General Public
// License.  This exception does not however invalidate any other
// reasons why the executable file might be covered by the GNU General
// Public License.

#ifndef OLENA_IO_PNM_READ_HXX_
# define OLENA_IO_PNM_READ_HXX_

# include "core/image.hh"
# include "core/iter.hh"
# include "value/bin.hh"
# include "value/int_u.hh"

# include "io/base.hh"

namespace oln {

  namespace io {

    namespace internal {

      template< typename T >
      bool pnm_read_header(std::istream& s, image2d<T>& im, char type)
      {
	// check magic
	if (s.get() != 'P' ) return false;
	if (s.get() != type) return false;
	if (s.get() != '\n') return false;

	// skip comments
	while (s.peek() == '#')
	  {
	    std::string line;
	    std::getline(s,line);
	  }

	// read size
	coord cols, rows;
	s >> cols >> rows;
	if (cols <= 0 || rows <= 0) return false;

	// resize image
	im = image2d<T>(rows, cols);

	// skip maxvalue
	if (s.get() != '\n') return false;
	if (type != '1' && type != '4') //FIXME: if (T==bin)
	  {
	    std::string line;
	    std::getline(s,line);
	  }
	return true;
      }

      template<>
      struct reader<ReadPnmPlain, image2d<bin> >
      {
	static const std::string& name()
	{ static const std::string _name("pnm/P1"); return _name;	}

	static bool knows_ext(const std::string& ext)
	{ return ext == "ppbm"; }

	static bool read(std::istream& in, image2d<bin>& im)
	{
	  if (!pnm_read_header(in, im, '1'))
	    return false;
	  image2d<bin>::iter it(im);
	  for ( it = begin; it != end; ++it )
	    {
	      unsigned char c;
	      do
		c = in.get();
	      while ((c != '0') && (c != '1'));
	      im[it] = (c == '0');
	    }
	  return true;
	}
      };

      template<>
      struct reader<ReadPnmPlain, image2d<int_u8> >
      {
	static const std::string& name()
	{ static const std::string _name("pnm/P2"); return _name;	}

	static bool knows_ext(const std::string& ext)
	{ return ext == "ppgm"; }

	static bool read(std::istream& in, image2d<int_u8>& im)
	{
	  if (!pnm_read_header(in, im, '2'))
	    return false;
	  image2d<int_u8>::iter it(im);
	  for ( it = begin; it != end; ++it )
	    {
	      int tmp;
	      in >> tmp;
	      im[it] = tmp;
	    }
	  return true;
	}
      };

      //FIXME: replace int_u<24> with rgb<int_u<8>>
      template<>
      struct reader<ReadPnmPlain, image2d<int_u<24> > >
      {
	static const std::string& name()
	{ static const std::string _name("pnm/P3"); return _name;	}

	static bool knows_ext(const std::string& ext)
	{ return ext == "pppm"; }

	static bool read(std::istream& in, image2d<int_u<24> >& im)
	{
	  if (!pnm_read_header(in, im, '3'))
	    return false;
	  image2d<int_u<24> >::iter it(im);
	  for ( it = begin; it != end; ++it )
	    {
	      int tmp;
	      in >> tmp; im[it] = tmp; im[it] = im[it] << 8;
	      in >> tmp; im[it] += tmp; im[it] = im[it] << 8;
	      in >> tmp; im[it] += tmp;
	      //FIXME: should be this with rgb<int_u8>
	      //  int tmp;
	      //  in >> tmp; im[it].r = tmp;
	      //  in >> tmp; im[it].g = tmp;
	      //  in >> tmp; im[it].b = tmp;
	    }
	  return true;
	}
      };

      template<>
      struct reader<ReadPnmRaw, image2d<bin> >
      {
	static const std::string& name()
	{ static const std::string _name("pnm/P4"); return _name;	}

	static bool knows_ext(const std::string& ext)
	{ return ext == "pbm"; }

	static bool read(std::istream& in, image2d<bin>& im)
	{
	  if (!pnm_read_header(in, im, '4'))
	    return false;
	  image2d<bin>::iter it(im);
	  coord cols = 0;
	  unsigned bits = 0;
	  unsigned char c = 0;
	  for (it = begin; it != end; ++it)
	    {
	      if (bits == 0)
		{
		  c = in.get();
		  bits = 8;
		}
	      im[it] = (c & (1 << --bits)) ? false : true;
	      if (++cols >= im.ncols())
		cols = bits = 0;
	    }
	  return true;
	}
      };

      template<>
      struct reader<ReadPnmRaw, image2d<int_u8> >
      {
	static const std::string& name()
	{ static const std::string _name("pnm/P5"); return _name;	}

	static bool knows_ext(const std::string& ext)
	{ return ext == "pgm"; }

	static bool read(std::istream& in, image2d<int_u8>& im)
	{
	  if (!pnm_read_header(in, im, '5'))
	    return false;
	  if (im.border())
	    for (coord j = 0; j < im.nrows(); ++j)
	      in.read((char*)&im(j,0), im.ncols());
	  else
	    in.read((char*)&im(0,0), im.ncols() * im.nrows());
	  return true;
	}
      };

      //FIXME: replace int_u<24> with rgb<int_u<8>>
      template<>
      struct reader<ReadPnmRaw, image2d< int_u<24> > >
      {
	static const std::string& name()
	{ static const std::string _name("pnm/P6"); return _name;	}

	static bool knows_ext(const std::string& ext)
	{ return ext == "ppm"; }

	static bool read(std::istream& in, image2d< int_u<24> >& im)
	{
	  if (!pnm_read_header(in, im, '6'))
	    return false;
	  image2d<int_u8>::iter it(im);
	  for ( it = begin; it != end; ++it )
	    {
	      int_u8::value_t tmp;
	      in.read((char *)&tmp, 1);
	      im[it] = tmp; im[it] = im[it] << 8;
	      in.read((char *)&tmp, 1);
	      im[it] += tmp; im[it] = im[it] << 8;
	      in.read((char *)&tmp, 1);
	      im[it] += tmp;
	      //FIXME: should be this with rgb<int_u8>
	      // if (im.border())
	      //   for (coord j = 0; j < im.nrows(); ++j)
	      //     in.read((char*)&im(j,0), 3 * im.ncols());
	      // else
	      //   in.read((char*)&im(0,0), 3 * im.ncols() * im.nrows());
	    }
	  return true;
	}
      };


    } // end of internal

  } // end of io

} // end of oln


#endif // OLENA_IO_PNM_READ_HXX_
