// Copyright 2001  EPITA Research and Development Laboratory
//
// This file is part of the Olena Library.  This library is free
// software; you can redistribute it and/or modify it under the terms
// of the GNU General Public License version 2 as published by the
// Free Software Foundation.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this library; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// As a special exception, you may use this file as part of a free
// software library without restriction.  Specifically, if other files
// instantiate templates or use macros or inline functions from this
// file, or you compile this file and link it with other files to
// produce an executable, this file does not by itself cause the
// resulting executable to be covered by the GNU General Public
// License.  This exception does not however invalidate any other
// reasons why the executable file might be covered by the GNU General
// Public License.

#ifndef OLENA_IO_PNM_WRITE_HXX_
# define OLENA_IO_PNM_WRITE_HXX_

# include "core/image.hh"
# include "core/iter.hh"
# include "value/bin.hh"
# include "value/int_u.hh"
# include "io/base.hh"

namespace oln {

  namespace io {

    namespace internal {

      using oln::internal::image2d_data;

      template< typename T >
      bool pnm_write_header(std::ostream& s, const image2d<T>& im, char type)
      {
	s.put('P'); s.put(type); s.put('\n');
	s << "# Creator: OLENA / Epita-LRDE" << std::endl
	  << im.ncols() << ' ' << im.nrows() << std::endl;
	if (type != '1' && type != '4') //FIXME: if (T==bin)
	  s << T::max() << std::endl;
	return true;
      }


      template<>
      struct writer<WritePnmPlain, image2d<bin> >
      {
	static const std::string& name()
	{ static const std::string _name("pnm/P1"); return _name; }

	static bool knows_ext(const std::string& ext)
	{ return ext == "ppbm"; }

	static bool write(std::ostream& out, const image2d<bin>& im)
	{
	  if (!pnm_write_header(out, im, '1'))
	    return false;
	  image2d<bin>::iter it(im);
	  coord stride = 0;
	  coord cols = 0;
	  for ( it = begin; it != end; ++it )
	    {
	      out.put(im[it] == true ? '0' : '1');
	      if (++cols >= im.ncols())
		{
		  out << std::endl;
		  cols = stride = 0;
		}
	      else
		if (++stride >= 70)
		  {
		    out << std::endl;
		    stride = 0;
		  }
	    }
	  return true;
	}
      };

      template<unsigned N>
      struct writer<WritePnmPlain, image2d< int_u<N> > >
      {
	static const std::string& name()
	{ static const std::string _name("pnm/P2"); return _name;	}

	static bool knows_ext(const std::string& ext)
	{ return ext == "ppgm"; }

	static bool write(std::ostream& out, const image2d< int_u<N> >& im)
	{
	  if (!pnm_write_header(out, im, '2'))
	    return false;
	  image2d<int_u<N> >::iter it(im);
	  coord stride = 0;
	  coord cols = 0;
	  for ( it = begin; it != end; ++it )
	    {
	      out << im[it];
	      if (++cols >= im.ncols())
		{
		  out << std::endl;
		  cols = stride = 0;
		}
	      else
		{
		  //FIXME: should be log(N)/log(2.f) instead of 3
		  // but it doesn't work for int_u8
		  stride += 3+1;
		  if (stride >= 70-3)
		    {
		      out << std::endl;
		      stride = 0;
		    }
		  else
		    out << ' ';
		}
	    }
	  return true;
	}
      };


      //FIXME: replace int_u<24> with rgb<int_u8>
      template<>
      struct writer<WritePnmPlain, image2d< int_u<24> > >
      {
#define __value_len	3   //FIXME: should be log(N)/log(2.f) to handle int_u<N>
#define __nb_component	3
#define __space		' '
#define __space_len	1
#define __separator	"  "
#define __separator_len	2
#define __stride_len	(__nb_component*__value_len+(__nb_component-1)*__space_len)

	static const std::string& name()
	{ static const std::string _name("pnm/P3"); return _name;	}

	static bool knows_ext(const std::string& ext)
	{ return ext == "pppm"; }

	static bool write(std::ostream& out, const image2d< int_u<24> >& im)
	{
	  image2d< int_u<8> > dummy(im.size());	  //FIXME: so T::max() is 255 in pnmheader
	  if (!pnm_write_header(out, dummy, '3')) // remove with rgb<int_u8>
	    return false;
	  image2d<int_u<24> >::iter it(im);
	  coord stride = 0;
	  coord cols = 0;
	  for ( it = begin; it != end; ++it )
	    {
	      out << int_u8((im[it] >> 16) & 0xFF) //FIXME: im[it].r
		  << __space
		  << int_u8((im[it] >> 8) & 0xFF) //FIXME: im[it].g
		  << __space
		  << int_u8(im[it] & 0xFF); //FIXME: im[it].b
	      if (++cols >= im.ncols())
		{
		  out << std::endl;
		  cols = stride = 0;
		}
	      else
		{
		  stride += __stride_len + __separator_len;
		  if (stride >= 70-__stride_len)
		    {
		      out << std::endl;
		      stride = 0;
		    }
		  else
		    out << __separator;
		}
	    }
	  return true;
	}
      };

      template<>
      struct writer<WritePnmRaw, image2d<bin> >
      {
	static const std::string& name()
	{ static const std::string _name("pnm/P4"); return _name;	}

	static bool knows_ext(const std::string& ext)
	{ return ext == "pbm"; }

	static bool write(std::ostream& out, const image2d<bin>& im)
	{
	  if (!pnm_write_header(out, im, '4'))
	    return false;
	  image2d<bin>::iter it(im);
	  coord cols = 0;
	  unsigned char c = 0;
	  coord stride = 0;
	  for ( it = begin; it != end; ++it )
	    {
	      c <<= 1;
	      c += (im[it] == true ? 0 : 1);
	      if (++cols >= im.ncols())
		{
		  out << c;
		  cols = stride = 0;
		}
	      else
		if (++stride >= 8)
		  {
		    out << c;
		    stride = 0;
		  }
	    }
	  if (stride > 0)
	    out << c;
	  return true;
	}
      };

      template<unsigned N>
      struct writer<WritePnmRaw, image2d< int_u<N> > >
      {
	static const std::string& name()
	{ static const std::string _name("pnm/P5"); return _name;	}

	static bool knows_ext(const std::string& ext)
	{ return ext == "pgm"; }

	static bool write(std::ostream& out, const image2d< int_u<N> >& im)
	{
	  if (!pnm_write_header(out, im, '5'))
	    return false;
	  if (im.border())
	    for (coord j = 0; j < im.nrows(); ++j)
	      out.write((char*)&im.data()->at(j,0),
			sizeof(int_u<N>::real_value_t) * im.ncols());
	  else
	    out.write((char*)&im.data()->at(0,0),
		      sizeof(int_u<N>::real_value_t) * im.ncols()*im.nrows());
	  return true;
	}
      };

      //FIXME: replace int_u<24> with rgb<int_u<8>>
      template<>
      struct writer<WritePnmRaw, image2d< int_u<24> > >
      {
	static const std::string& name()
	{ static const std::string _name("pnm/P6"); return _name;	}

	static bool knows_ext(const std::string& ext)
	{ return ext == "ppm"; }

	static bool write(std::ostream& out, const image2d< int_u<24> >& im)
	{
	  image2d< int_u<8> > dummy(im.size());	  //FIXME: so T::max() is 255 in pnmheader
	  if (!pnm_write_header(out, dummy, '6')) // remove with rgb<int_u8>
	    return false;
	  image2d<int_u<24> >::iter it(im);
	  for ( it = begin; it != end; ++it )
	    {
	      int_u8::value_t r = (real_value(im[it]) >> 16) & 0xFF;
	      out.write((char*)&r, 1);
	      int_u8::value_t g = (real_value(im[it]) >> 8) & 0xFF;
	      out.write((char*)&g, 1);
	      int_u8::value_t b = real_value(im[it]) & 0xFF;
	      out.write((char*)&b, 1);
	    }
	  //FIXME: should be this with rgb<int_u8>
	  //  	  if (im.border())
	  //  	    for (coord j = 0; j < im.nrows(); ++j)
	  //  	      out.write(&im(j,0), 3 * im.ncols());
	  //  	  else
	  //  	    out.write(&im(0,0), 3 * im.ncols() * im.nrows());
	  return true;
	}
      };


    } // end of internal

  } // end of io

} // end of oln


#endif // OLENA_IO_PNM_WRITE_HXX_
